"""Helper classes and function to iterate through BIDS and Clinica dataset."""

import glob
import os
import re
from typing import Generator


class BIDSDirectory:
    """Utility to query BIDS directories."""

    has_session: bool = False
    has_sites: bool = False
    sites: list[str] | None = None

    def __init__(self, dataset_path: str):
        """Initialize a directory.

        Args:
        dataset (str): Dataset directory name
        root_dir (str): Root directory for all datasets
        """
        self.dataset_path = dataset_path

        self.has_session = len(glob.glob("sub-*/ses-*", root_dir=self.dataset_path)) > 0

    def get_subjects(self) -> list[str]:
        """Retrieve list of all subjects.

        Returns:
            list[str]: list of all subjects
        """
        return glob.glob("sub-*", root_dir=self.dataset_path)

    def get_sessions(self, sub_id: str) -> list[str]:
        """Retrieve list of sessions.

        Args:
            sub_id (str): Subject id (sub-???)
        Returns:
            list[str]: list of sessions
        """
        return glob.glob("ses-*", root_dir=os.path.join(self.dataset_path, sub_id))

    def walk(self) -> Generator[tuple[str, ...], None, None]:
        """Iterate over every subjects, session and generation of the dataset.

        Yields:
            Generator[tuple[str, ...], None, None]: sub_id, ses_id
        """
        for sub in self.get_subjects():
            if self.has_session:
                for ses in self.get_sessions(sub):
                    yield sub, ses
            else:
                yield (sub,)

    def get_all_t1w(self, sub_id: str, ses_id: str | None = None) -> list[str]:
        """Return path to all T1w volumes.

        Args:
            sub_id (str): Subject id (sub-???)
            ses_id (str, optional): Session id (ses-???). Defaults to None.

        Returns:
            list[str] : Path to all T1w
        """
        if ses_id is not None:
            return glob.glob(
                os.path.join(self.dataset_path, sub_id, ses_id, "anat", "*T1w.nii.gz")
            )
        return glob.glob(os.path.join(self.dataset_path, sub_id, "anat", "*T1w.nii.gz"))

    def extract_sub_ses(self, path: str) -> tuple[str | None, str | None]:
        """Extract subject and session identifier from path.

        Args:
            path (str): path to NifTi file

        Returns:
            tuple[str | None, str | None]: sub-id, ses-id
        """
        sub = None
        ses = None

        req_sub = r".*(sub-[\dA-Za-z]*).*"
        req_ses = r".*(ses-[\dA-Za-z]*).*"
        match_sub = re.match(req_sub, path)
        match_ses = re.match(req_ses, path)

        if match_sub is not None:
            groups = match_sub.groups()
            sub = groups[0]

        if match_ses is not None:
            groups = match_ses.groups()
            ses = groups[0]
        return sub, ses

    def __len__(self):
        """Return len of available T1w volumes."""
        return len(list(self.walk()))


class ClinicaDirectory(BIDSDirectory):
    """Utility class to query Clinica processed directory."""

    def __init__(self, dataset_path: str):
        """Initialize a Clinica directory.

        Args:
        dataset (str): Dataset directory name
        root_dir (str): Root directory for all datasets
        """
        super().__init__(dataset_path)
        self.dataset_path = os.path.join(self.dataset_path, "subjects")

        # Clinica datasets necessarily have session because clinica will not process
        # a bids datset without it
        self.has_session = len(glob.glob("sub-*/ses-*", root_dir=self.dataset_path)) > 0
        assert self.has_session, "Clinica datasets should have sessions"

    def get_all_t1w(self, sub_id: str, ses_id: str = "ses-1") -> list[str]:
        """Return path to T1w volume.

        Args:
            sub_id (str): Subject id (sub-???)
            ses_id (str, optional): Session id (ses-???). Defaults to "ses-1".
            gen_id : Ignores
        Returns:
            str: Path to T1w
        """
        return glob.glob(
            os.path.join(
                self.dataset_path,
                sub_id,
                ses_id,
                "t1_linear",
                "*space-MNI152NLin2009cSym_res-1x1x1_T1w.nii.gz",
            )
        )



import subprocess

import pytest

from neuro_ix_tools.processing.apptainer import ApptainerEnv


@pytest.fixture()
def fs_container():
    return ApptainerEnv.freesurfer()


def test_apptainer_bind(fs_container):
    fs_container.bind("origin_dir/ds", "destination_dir/ds/mnt")
    command = fs_container.compile()

    assert "--bind origin_dir/ds:destination_dir/ds/mnt" in command


def test_apptainer_add_command(fs_container):
    fs_container.add_command("echo 'here'")
    command = fs_container.compile()

    assert "echo 'here'" in command


def test_apptainer_run(fs_container):
    fs_container.bind("tests/data/apptainer_test/", "/mnt/test_dir")
    fs_container.add_command("cat /mnt/test_dir/test_file.txt")
    command = fs_container.compile()

    out = subprocess.check_output(command, shell=True)
    out_str = out.decode("utf-8")
    assert out_str == "Check Apptainer bind and command"



"""Module defining abstract command class."""

import abc


class Command(abc.ABC):
    """Basic Command wrapper definition."""

    cmd: str

    def compile(self) -> str:
        """Create the final command to execute.

        Returns:
            str: Final command
        """
        return self.cmd



"""Module retrieving necessary env variable and storing them as constants."""

import logging
import os

import dotenv
from platformdirs import user_config_dir

dotenv.load_dotenv()

CONFIG_FILE = os.path.join(user_config_dir("neuro-ix-tools"), "config.env")

if os.path.exists(".env"):
    dotenv.load_dotenv()
else:
    if os.path.exists(CONFIG_FILE):
        dotenv.load_dotenv(CONFIG_FILE)
    else:
        logging.warning("No config found at %s",CONFIG_FILE)

PATH_FREESURFER_SIF = os.getenv("PATH_FREESURFER_SIF")
DEFAULT_SLURM_ACCOUNT = os.getenv("DEFAULT_SLURM_ACCOUNT")
FREESURFER_LOGS = os.getenv("FREESURFER_LOGS", "~/freesurfer_pipeline/logs/")



import os
import re
import shutil

import pytest
from simple_slurm import Slurm

from neuro_ix_tools.bids import BIDSDirectory
from neuro_ix_tools.processing.freesurfer import (
    ApptainerFreesurfer,
    FreeSurferReconAll,
    cpy_all,
    cpy_cortical_stats_only,
    create_subject_directories,
    slurm_freesurfer_all_results,
    slurm_freesurfer_cortical_stats,
)


def test_freesurfer_recon_all():
    recon_all = FreeSurferReconAll(
        subject="sub-111", path="path/to/sub-111", subject_dir="/output/for/sub"
    )
    command = recon_all.compile()
    assert command == "recon-all -all -s sub-111 -i path/to/sub-111 -sd /output/for/sub"

    recon_all = FreeSurferReconAll(subject="sub-111", path="path/to/sub-111")
    command = recon_all.compile()
    assert command == "recon-all -all -s sub-111 -i path/to/sub-111"


@pytest.fixture()
def apptainer_fs():
    return ApptainerFreesurfer(
        "sub-000103",
        "ses-standard",
        BIDSDirectory("tests/data/bids_sub_ses"),
        "test_outputs",
    )


@pytest.fixture()
def ds():
    return BIDSDirectory("tests/data/bids_sub_ses")


def test_apptainer_freesurfer_compile(apptainer_fs: ApptainerFreesurfer):
    command = apptainer_fs.compile()
    fs_command = apptainer_fs.freesurfer_cmd.compile()
    assert fs_command in command
    assert (
        "--bind tests/data/bids_sub_ses/sub-000103/ses-standard/anat:/data" in command
    )
    assert "--bind test_outputs:/tmp" in command


def test_apptainer_freesurfer_to_slurm(apptainer_fs: ApptainerFreesurfer):
    job = apptainer_fs.to_slurm()
    command = apptainer_fs.compile()

    assert isinstance(job, Slurm)
    assert command in str(job)


def test_create_subject_directories(ds):
    sub_dir = create_subject_directories("sub-000103", "ses-standard", ds)

    assert os.path.exists("tests/data/bids_sub_ses/derivatives")
    assert os.path.exists("tests/data/bids_sub_ses/derivatives/sub-000103")
    assert os.path.exists("tests/data/bids_sub_ses/derivatives/sub-000103/ses-standard")
    assert sub_dir == "tests/data/bids_sub_ses/derivatives/sub-000103/ses-standard"

    shutil.rmtree("tests/data/bids_sub_ses/derivatives")


def test_cpy_cortical_stats_only(apptainer_fs: ApptainerFreesurfer, ds):
    job = apptainer_fs.to_slurm()
    create_subject_directories("sub-000103", "ses-standard", ds)

    cpy_cortical_stats_only(
        "tests/data/bids_sub_ses/derivatives/sub-000103/ses-standard", "test_fsdir", job
    )

    assert os.path.exists(
        "tests/data/bids_sub_ses/derivatives/sub-000103/ses-standard/stats"
    )
    match = re.search(
        r"mv \\\$SLURM_TMPDIR\/test_fsdir\/stats\/rh\.aparc\.stats.*tests\/data\/bids_sub_ses\/derivatives\/sub-000103\/ses-standard\/stats\/rh\.aparc\.stats",
        str(job),
    )
    assert match is not None

    shutil.rmtree("tests/data/bids_sub_ses/derivatives")


def test_cpy_cortical_stats_only_auto_rm(apptainer_fs: ApptainerFreesurfer, ds):
    job = apptainer_fs.to_slurm()
    create_subject_directories("sub-000103", "ses-standard", ds)

    os.makedirs("tests/data/bids_sub_ses/derivatives/sub-000103/ses-standard/stats")
    cpy_cortical_stats_only(
        "tests/data/bids_sub_ses/derivatives/sub-000103/ses-standard", "test_fsdir", job
    )

    assert os.path.exists(
        "tests/data/bids_sub_ses/derivatives/sub-000103/ses-standard/stats"
    )
    match = re.search(
        r"mv \\\$SLURM_TMPDIR\/test_fsdir\/stats\/rh\.aparc\.stats.*tests\/data\/bids_sub_ses\/derivatives\/sub-000103\/ses-standard\/stats\/rh\.aparc\.stats",
        str(job),
    )
    assert match is not None

    shutil.rmtree("tests/data/bids_sub_ses/derivatives")


def test_cpy_all(apptainer_fs: ApptainerFreesurfer, ds):
    job = apptainer_fs.to_slurm()
    create_subject_directories("sub-000103", "ses-standard", ds)

    cpy_all(
        "tests/data/bids_sub_ses/derivatives/sub-000103/ses-standard", "test_fsdir", job
    )

    match = re.search(
        r"mv \\\$SLURM_TMPDIR\/test_fsdir.*tests\/data\/bids_sub_ses\/derivatives\/sub-000103\/ses-standard\/",
        str(job),
    )
    assert match is not None

    shutil.rmtree("tests/data/bids_sub_ses/derivatives")


def test_slurm_freesurfer_cortical_stats(ds):
    job = slurm_freesurfer_cortical_stats("sub-000103", "ses-standard", ds)
    assert isinstance(job, Slurm)


def test_slurm_freesurfer_all_results(ds):
    job = slurm_freesurfer_all_results("sub-000103", "ses-standard", ds)
    assert isinstance(job, Slurm)



"""Module defining click commands."""

import os

import click

from neuro_ix_tools import config
from neuro_ix_tools.bids import BIDSDirectory, ClinicaDirectory
from neuro_ix_tools.processing.freesurfer import (
    slurm_freesurfer_all_results,
    slurm_freesurfer_cortical_stats,
)


@click.group("freesurfer")
def freesurfer_cli():
    """Freesurfer related commands."""


@freesurfer_cli.command()
@click.option(
    "-b",
    "--bids-dataset",
    help="absolute path to BIDS compliant dataset",
    type=str,
    default=None,
)
@click.option(
    "-c",
    "--clinica-dataset",
    help="absolute path to CLINICA compliant dataset",
    type=str,
    default=None,
)
@click.option(
    "--cortical-stats", help="only keep cortical stats", type=bool, is_flag=True
)
@click.option(
    "--start-from",
    type=int,
    default=0,
    help="Specify at which volume index should the processing start",
)
def recon_all(
    bids_dataset: str, clinica_dataset: str, cortical_stats: bool, start_from: int
):
    """Compute freesurfer recon-all."""
    assert bids_dataset is not None or clinica_dataset is not None, (
        "One dataset path should be specified"
    )
    assert bids_dataset is None or clinica_dataset is None, (
        "Only one dataset path should be specified"
    )

    if not os.path.exists(config.FREESURFER_LOGS):
        os.makedirs(config.FREESURFER_LOGS)

    if bids_dataset is not None:
        ds = BIDSDirectory(bids_dataset)
    else:
        ds = ClinicaDirectory(clinica_dataset)

    if cortical_stats:
        slurm_cmd = slurm_freesurfer_cortical_stats
    else:
        slurm_cmd = slurm_freesurfer_all_results
    sub_ses = list(ds.walk())
    for sub, ses in sub_ses[start_from:1000]:
        slurm_cmd(sub, ses, ds).sbatch()

    if len(sub_ses) > 1000:
        click.secho(
            (
                f"Dataset contain {len(sub_ses)} but only the first 1000 can be processed"
                "due to limits on slurm cluster jobs"
            ),
            color="red",
            bold=True,
        )
        click.secho(
            (
                "Once the jobs are finished, launch the same command with the argument"
                " : --start-from {start_from + 1000}"
            ),
            bold=True,
        )



"""Module defining Slurm related utility."""

from simple_slurm import Slurm


def get_apptainer_job(
    mem: str, time: str, account: str, output: str, cpus: int = 1
) -> Slurm:
    """Create a Slurm job with apptainer activated.

    Args:
        mem (str): memory to allocate
        time (str): duration of task
        account (str): slurm account for ressources
        output (str): output path
        cpus (int, optional): number of cpu to allocate. Defaults to 1.

    Returns:
        Slurm: job with apptainer module activated
    """
    job = Slurm(
        cpus_per_task=cpus,
        mem=mem,
        account=account,
        time=time,
        output=output,
    )
    job.add_cmd("module load apptainer")

    return job



import pytest

from neuro_ix_tools.bids import BIDSDirectory, ClinicaDirectory


# TESTS ON SUB ONLY DATASET
@pytest.fixture()
def bids_sub():
    return BIDSDirectory("tests/data/bids_sub")


def test_bids_sub_get_subjects(bids_sub):
    subjects = bids_sub.get_subjects()

    assert len(subjects) == 2


def test_bids_sub_get_sessions(bids_sub):
    session = bids_sub.get_sessions("sub-000103")

    assert not bids_sub.has_session
    assert len(session) == 0


def test_bids_sub_walk(bids_sub):
    walked = list(bids_sub.walk())

    assert len(walked) == 2
    assert len(walked[0]) == 1, "Walk seems to report sessions"
    assert len(walked) == len(bids_sub)


def test_bids_sub_get_all_t1w(bids_sub):
    t1ws = list(bids_sub.get_all_t1w("sub-000103"))

    assert len(t1ws) == 1
    assert "T1w.nii.gz" in t1ws[0]


def test_bids_sub_extract_sub_ses(bids_sub):
    t1ws = list(bids_sub.get_all_t1w("sub-000103"))
    sub, ses = bids_sub.extract_sub_ses(t1ws[0])
    assert sub == "sub-000103"
    assert ses is None


# TESTS ON SUB AND SES DATASET
@pytest.fixture()
def bids_sub_ses():
    return BIDSDirectory("tests/data/bids_sub_ses")


def test_bids_sub_ses_get_subjects(bids_sub_ses):
    subjects = bids_sub_ses.get_subjects()

    assert len(subjects) == 2


def test_bids_sub_ses_get_sessions(bids_sub_ses):
    session = bids_sub_ses.get_sessions("sub-000103")

    assert bids_sub_ses.has_session
    assert len(session) == 2


def test_bids_sub_ses_walk(bids_sub_ses):
    walked = list(bids_sub_ses.walk())

    assert len(walked) == 3
    assert len(walked[0]) == 2, "Walk did not report sessions"
    assert len(walked) == len(bids_sub_ses)


def test_bids_sub_ses_get_all_t1w(bids_sub_ses):
    t1ws = list(bids_sub_ses.get_all_t1w("sub-000103", "ses-headmotion2"))

    assert len(t1ws) == 1
    assert "T1w.nii.gz" in t1ws[0]


def test_bids_sub_ses_extract_sub_ses(bids_sub_ses):
    t1ws = list(bids_sub_ses.get_all_t1w("sub-000103", "ses-standard"))
    sub, ses = bids_sub_ses.extract_sub_ses(t1ws[0])
    assert sub == "sub-000103"
    assert ses == "ses-standard"


# TESTS ON CLINICA DATASET
@pytest.fixture()
def clinica():
    return ClinicaDirectory("tests/data/clinica")


def test_clinica_get_subjects(clinica):
    subjects = clinica.get_subjects()

    assert len(subjects) == 2


def test_clinica_get_sessions(clinica):
    session = clinica.get_sessions("sub-000103")

    assert len(session) == 2


def test_clinica_walk(clinica):
    walked = list(clinica.walk())

    assert len(walked) == 3
    assert len(walked[0]) == 2, "Walk did not report sessions"
    assert len(walked) == len(clinica)


def test_clinica_get_all_t1w(clinica):
    t1ws = list(clinica.get_all_t1w("sub-000103", "ses-headmotion2"))

    assert len(t1ws) == 1
    assert "T1w.nii.gz" in t1ws[0]


def test_clinica_extract_sub_ses(clinica):
    t1ws = list(clinica.get_all_t1w("sub-000103", "ses-standard"))
    sub, ses = clinica.extract_sub_ses(t1ws[0])
    assert sub == "sub-000103"
    assert ses == "ses-standard"



"""Module containing the logic to automate freesurfer jobs on Narval."""

import logging
import os
import shutil

from simple_slurm import Slurm

from neuro_ix_tools import config
from neuro_ix_tools.bids import BIDSDirectory
from neuro_ix_tools.processing.apptainer import ApptainerEnv
from neuro_ix_tools.processing.base_cmd import Command
from neuro_ix_tools.processing.slurm import get_apptainer_job


class FreeSurferReconAll(Command):
    """Object to wrap FreeSurfer's recon-all command."""

    def __init__(self, subject: str, path: str, subject_dir: str | None = None):
        """Initialize FreeSurfer recon-all command.

        Args:
        subject (str): Subject directory output name
        path (str): Path to the subject T1w volume
        subject_dir (str, optional): Directory for outputs. Defaults to None.
        """
        self.cmd = f"recon-all -all -s {subject} -i {path}"
        if subject_dir:
            self.cmd += f" -sd {subject_dir}"


class ApptainerFreesurfer(Command):
    """Object to control a freesurfer command inside an apptainer container."""

    def __init__(
        self, sub_id: str, ses_id: str, dataset: BIDSDirectory, subject_output: str
    ):
        """Initialize a Freesurfer command in an Apptainer container.

        Args:
            sub_id (str): subject to process
            ses_id (str): session to process
            dataset (BIDSDirectory): dataset containing the volume
            subject_output (str): final folder to store freesurfer outputs
        """
        # the directory where freesurfer will store results(needs to be unique)
        self.fs_dir = sub_id + ses_id
        self.sub_id = sub_id
        self.ses_id = ses_id

        t1_for_sub = dataset.get_all_t1w(sub_id, ses_id)
        assert len(t1_for_sub) == 1, (
            f"Wrong number of volumes for subject {sub_id} at session {ses_id}."
            f" Expected 1, found {len(t1_for_sub)}"
        )
        orig_volume_path = t1_for_sub[0]

        self.input_root = os.path.dirname(orig_volume_path)
        self.env_volume_path = os.path.join("/data", os.path.basename(orig_volume_path))
        # Creating an Apptainer Command to use freesurfer with apptainer
        self.apptainer_env = ApptainerEnv.freesurfer()
        self.apptainer_env.bind(subject_output, "/tmp")
        self.apptainer_env.bind(self.input_root, "/data")

        self.freesurfer_cmd = FreeSurferReconAll(
            self.fs_dir, self.env_volume_path, subject_dir="/tmp"
        )
        self.apptainer_env.add_command(self.freesurfer_cmd.compile())

    def compile(self) -> str:
        """Return command to execute as a string."""
        return self.apptainer_env.compile()

    def to_slurm(self) -> Slurm:
        """Wrap command in a Slurm job.

        Returns:
            Slurm: Ready to use Slurm job
        """
        output_slurm = os.path.join(
            config.FREESURFER_LOGS, f"freesurfer_{self.sub_id}_{self.ses_id}.%j.out"
        )

        assert config.DEFAULT_SLURM_ACCOUNT is not None, (
            "Missing default Slurm account (define env variable $DEFAULT_SLURM_ACCOUNT)"
        )
        job = get_apptainer_job(
            mem="30G",
            time="10:00:00",
            account=config.DEFAULT_SLURM_ACCOUNT,
            cpus=1,
            output=output_slurm,
        )

        job.add_cmd(self.compile())
        return job


def create_subject_directories(sub_id: str, ses_id: str, dataset: BIDSDirectory) -> str:
    """Create necessary output sub-folder structure in a `derivatives` folder.

    Args:
        sub_id (str): subject id
        ses_id (str): session id
        dataset (BIDSDirectory): dataset containing the volume

    Returns:
        str: path to the output dir
    """
    derivative_dir = os.path.join(dataset.dataset_path, "derivatives")
    os.makedirs(derivative_dir, exist_ok=True)
    sub_dir = os.path.join(derivative_dir, sub_id, ses_id)
    os.makedirs(sub_dir, exist_ok=True)
    return sub_dir


def cpy_cortical_stats_only(sub_dir: str, fs_dir: str, job: Slurm):
    """Modify job to only store freesurfer's stats files.

    It also creates a `stats` folder in the process
    Args:
        sub_dir (str): subject derivative output directory
        fs_dir (str): freesurfer output directory
            (Contains subject stats files)
        job (Slurm): job to modify
    """
    stat_dir = os.path.join(sub_dir, "stats")
    if os.path.exists(stat_dir):
        shutil.rmtree(stat_dir)
    os.makedirs(stat_dir)

    job.add_cmd(
        f"mv $SLURM_TMPDIR/{fs_dir}/stats/rh.aparc.stats \
              {os.path.join(stat_dir, 'rh.aparc.stats')}"
    )
    job.add_cmd(
        f"mv $SLURM_TMPDIR/{fs_dir}/stats/lh.aparc.stats \
              {os.path.join(stat_dir, 'lh.aparc.stats')}"
    )


def cpy_all(sub_dir: str, fs_dir: str, job: Slurm):
    """Modify job to copy freesurfer's output files.

    Args:
        sub_dir (str): subject derivative output directory
        fs_dir (str): freesurfer output directory
            (Contains subject stats files)
        job (Slurm): job to modify
    """
    job.add_cmd(
        f"mv $SLURM_TMPDIR/{fs_dir} \
              {sub_dir}/"
    )


def slurm_freesurfer_cortical_stats(
    sub_id: str, ses_id: str, dataset: BIDSDirectory
) -> Slurm:
    """Launch Slurm job to process one subject with freesurfer.

    Only keep cortical thickness stats.

    Args:
        sub_id (str): identifier of subject (with 'sub-')
        ses_id (str): identifier of session (with 'ses-')
        dataset (BIDSDirectoryDataset) : dataset object to retrieve pathes

    Returns:
        int: sbatch launch code
    """
    logging.info("Processing subject : %s", sub_id)

    # Creating an Apptainer Command to use freesurfer with apptainer
    apptainer_env = ApptainerFreesurfer(sub_id, ses_id, dataset, "$SLURM_TMPDIR")

    job = apptainer_env.to_slurm()

    # Create necessary directories
    sub_dir = create_subject_directories(sub_id, ses_id, dataset)

    # Copy cortical stats only
    cpy_cortical_stats_only(sub_dir, apptainer_env.fs_dir, job)

    return job


def slurm_freesurfer_all_results(
    sub_id: str, ses_id: str, dataset: BIDSDirectory
) -> Slurm:
    """Launch Slurm job to process one subject with freesurfer.

    Store every output files.

    Args:
        sub_id (str): identifier of subject (with 'sub-')
        ses_id (str): identifier of session (with 'ses-')
        dataset (BIDSDirectoryDataset) : dataset object to retrieve pathes

    Returns:
        int: sbatch launch code
    """
    logging.info("Processing subject : %s, session : %s", sub_id, ses_id)

    # Creating an Apptainer Command to use freesurfer with apptainer
    apptainer_env = ApptainerFreesurfer(sub_id, ses_id, dataset, "$SLURM_TMPDIR")

    job = apptainer_env.to_slurm()

    # Create necessary directories
    sub_dir = create_subject_directories(sub_id, ses_id, dataset)

    # Copy cortical stats only
    cpy_all(sub_dir, apptainer_env.fs_dir, job)

    return job
